/*
 * Jake Moilanen <moilanen@austin.ibm.com>
 * Copyright (C) 2006 IBM
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; version 2 of the
 * License.
 *
 */

#include <linux/pci.h>
#include <linux/irq.h>
#include <asm/rtas.h>
#include <asm/hw_irq.h>
#include <asm/ppc-pci.h>

int query_interrupt(struct pci_dev * pdev, int num_intrs)
{
	struct device_node * dn;
	struct pci_dn *pdn;
	int query_token = rtas_token("ibm,query-interrupt-source-number");
	unsigned int virq;
	unsigned int addr;
	unsigned long buid = -1;
	u32 *reg;
	int devfn;
	int busno;
	int reglen;
	int ret[3];
	int dummy;
	int rc;
	int i;

	dn = pci_device_to_OF_node(pdev);

	if (!of_find_property(dn, "ibm,req#msi", &dummy))
		return -ENOENT;

	reg = (u32 *) get_property(dn, "reg", &reglen);
	if (reg == NULL || reglen < 20)
		return -ENXIO;

	devfn = (reg[0] >> 8) & 0xff;
	busno = (reg[0] >> 16) & 0xff;

	pdn = PCI_DN(dn);
	if (!pdn)
		return -ENXIO;

	buid = pdn->phb->buid;
	addr = (busno << 16) | (devfn << 8);

	if (dn->intrs != dn->lsi_intrs)
		kfree(dn->intrs);

	dn->intrs = kmalloc(dn->n_intrs * sizeof(*(dn->intrs)), GFP_KERNEL);
	if (!dn->intrs) {
		printk(KERN_WARNING "query_interrupt: can't allocate space\n");
		return -ENOMEM;
	}

	for (i = 0; i < num_intrs; i++) {
		rc = rtas_call(query_token, 4, 3, ret, addr,
				       buid >> 32, buid & 0xffffffff, i);

		if (!rc) {
			virq = virt_irq_create_mapping(ret[0]);

			dn->intrs[i].line = irq_offset_up(virq);
			dn->intrs[i].sense = ret[1];
		} else {
			printk(KERN_WARNING "error[%d]: query-interrupt-source-number for %s\n",
			       rc, dn->name);
		}
	}

	return 0;
}

int rtas_enable_msi(struct pci_dev* pdev)
{
	int seq_num = 1;
	int rc;
	int devfn;
	int busno;
	u32 *reg;
	int reglen;
	int ret[3];
	unsigned int addr;
	unsigned long buid = -1;
	unsigned long wait_time;
	struct device_node * dn;
	struct pci_dn *pdn;

	dn = pci_device_to_OF_node(pdev);

	if (!of_find_property(dn, "ibm,req#msi", NULL))
		return -ENOENT;

	reg = (u32 *) get_property(dn, "reg", &reglen);
	if (reg == NULL || reglen < 20)
		return -ENXIO;

	devfn = (reg[0] >> 8) & 0xff;
	busno = (reg[0] >> 16) & 0xff;

	pdn = PCI_DN(dn);
	if (!pdn)
		return -ENXIO;

	buid = pdn->phb->buid;
	addr = (busno << 16) | (devfn << 8);

	while (1) {
		rc = rtas_call(rtas_token("ibm,change-msi"), 6, 4, ret, addr,
			       buid >> 32, buid & 0xffffffff,
			       3, 1, seq_num);

		/* if the new interface fails, try old API */
		if (rc == -3) {
			rc = rtas_call(rtas_token("ibm,change-msi"), 6, 3, ret, addr,
			       buid >> 32, buid & 0xffffffff,
			       1, 1, seq_num);
		}


		if (!rc)
			break;
		else if (rc == RTAS_BUSY)
			udelay(1);
		else if (rtas_is_extended_busy(rc)) {
			wait_time = rtas_extended_busy_delay_time(rc);
			udelay(wait_time * 1000);
		} else {
			printk(KERN_WARNING "error[%d]: getting the number of "
			       "MSI interrupts for %s\n", rc, dn->name);
			return -EIO;
		}

		seq_num = ret[1];
	}

	/* Return if there's no MSI interrupts */
	if (!ret[0]) {
		printk(KERN_WARNING "rtas_enable_msi: No interrupts assigned to slot\n");
		return -ENOENT;
	}

	dn->n_intrs = ret[0];

	rc = query_interrupt(pdev, dn->n_intrs);
	if (rc)
		return rc;

	/* Just give the first vector out for now */
	pdev->irq = dn->intrs[0].line;

	return 0;
}

void rtas_disable_msi(struct pci_dev* pdev)
{
	int seq_num = 1;
	struct device_node * dn;
	struct pci_dn *pdn;
	int rc;
	int devfn;
	int busno;
	u32 *reg;
	int reglen;
	int ret[3];
	int dummy;
	unsigned int addr;
	unsigned long buid = -1;
	unsigned long wait_time;

	dn = pci_device_to_OF_node(pdev);

	if (!of_find_property(dn, "ibm,req#msi", &dummy))
		return;

	reg = (u32 *) get_property(dn, "reg", &reglen);
	if (reg == NULL || reglen < 20)
		return;

	devfn = (reg[0] >> 8) & 0xff;
	busno = (reg[0] >> 16) & 0xff;

	pdn = PCI_DN(dn);
	if (!pdn)
		return;

	buid = pdn->phb->buid;
	addr = (busno << 16) | (devfn << 8);

	while (1) {
		rc = rtas_call(rtas_token("ibm,change-msi"), 6, 4, ret, addr,
			       buid >> 32, buid & 0xffffffff,
			       3, 0, seq_num);

		/* if the new interface fails, try old API */
		if (rc == -3) {
			rc = rtas_call(rtas_token("ibm,change-msi"), 6, 3, ret, addr,
				       buid >> 32, buid & 0xffffffff,
				       1, 0, seq_num);
		}

		if (!rc)
			break;
		else if (rc == RTAS_BUSY)
			udelay(1);
		else if (rtas_is_extended_busy(rc)) {
			wait_time = rtas_extended_busy_delay_time(rc);
			udelay(wait_time * 1000);
		} else {
			printk(KERN_WARNING "error[%d]: getting the number of "
			       "MSI interrupts for %s\n", rc, dn->name);
			return;
		}

		seq_num = ret[1];
	}

	rc = query_interrupt(pdev, 1);
	if (rc)
		return;

	dn->n_intrs = 1;

	pdev->irq = dn->intrs[0].line;
}
