#ifndef _LINUX_PAGE_STATES_H
#define _LINUX_PAGE_STATES_H

/*
 * include/linux/page-states.h
 *
 * (C) Copyright IBM Corp. 2005, 2006
 *
 * Authors: Martin Schwidefsky <schwidefsky@de.ibm.com>
 *          Hubertus Franke <frankeh@watson.ibm.com>
 *          Himanshu Raj <rhim@cc.gatech.edu>
 */

#if defined(CONFIG_PAGE_STATES)
#include <asm/page-states.h>
#else

/* Guest page hinting architecture primitives:
 * - page_host_discards:
 *     Indicates whether the host system discards guest pages or not.
 * - page_set_unused:
 *     Indicates to the host that the page content is of no interest
 *     to the guest. The host can "forget" the page content and replace
 *     it with a page containing zeroes.
 * - page_set_stable:
 *     Indicate to the host that the page content is needed by the guest.
 * - page_set_volatile:
 *     Make the page discardable by the host. Instead of writing the
 *     page to the hosts swap device, the host can remove the page.
 *     A guest that accesses such a discarded page gets a special
 *     discard fault.
 * - page_set_stable_if_present:
 *     The page state is set to stable if the page has not been discarded
 *     by the host. The check and the state change have to be done
 *     atomically.
 * - page_discarded:
 *     Returns true if the page has been discarded by the host.
 */

#define page_host_discards()			(0)
#define page_set_unused(_page,_order)		do { } while (0)
#define page_set_stable(_page,_order)		do { } while (0)
#define page_set_volatile(_page,_writable)	do { } while (0)
#define page_set_stable_if_present(_page)	(1)
#define page_discarded(_page)			(0)

#endif

/*
 * Common guest page hinting functions. The compiler will turn them
 * into nop's if page_host_discards as primitive is defined as (0):
 * - page_make_stable:
 *     Tries to make a page stable. This operation can fail if the
 *     host has discarded a page. The function returns != 0 if the
 *     page could not be made stable.
 * - page_make_volatile:
 *     Tries to make a page volatile. There are a number of conditions
 *     that prevent a page from becoming volatile. If at least one
 *     is true the function does nothing. See mm/page-states.c for
 *     details.
 * - page_discard:
 *     Removes a discarded page from the system. The page is removed
 *     from the LRU list and the radix tree of its mapping.
 *     page_discard uses page_unmap_all to remove all page table
 *     entries for a page.
 * - page_check_writable:
 *     Checks if the page states needs to be adapted because a new
 *     writable page table entry refering to the page is established.
 * - page_reset_writable:
 *     Resets the page state after the last writable page table entry
 *     refering to the page has been removed.
 * - page_shrink_discards:
 *     Frees all pages that free_hot_cold_page has put on the list of
 *     discarded pages.
 */
extern void page_unmap_all(struct page *page);
extern void page_discard(struct page *page);
extern unsigned long page_shrink_discards(void);

static inline int page_make_stable(struct page *page)
{
	extern int  __page_make_stable(struct page *);
	if (!page_host_discards())
		return 1;
	return __page_make_stable(page);
}

static inline void page_make_volatile(struct page *page, unsigned int offset)
{
	extern void __page_make_volatile(struct page *, unsigned int offset);
	if (!page_host_discards())
		return;
	if (likely(!test_bit(PG_discarded, &page->flags)))
		__page_make_volatile(page, offset);
}

static inline void page_check_writable(struct page *page, pte_t pte,
				       unsigned int offset)
{
	extern void __page_check_writable(struct page *, pte_t, unsigned int);
	if (!page_host_discards() || !pte_write(pte) ||
	    test_bit(PG_writable, &page->flags))
		return;
	__page_check_writable(page, pte, offset);
}

static inline void page_reset_writable(struct page *page)
{
	extern void __page_reset_writable(struct page *);
	if (!page_host_discards() || !test_bit(PG_writable, &page->flags))
		return;
	__page_reset_writable(page);
}

#endif /* _LINUX_PAGE_STATES_H */
