/*
 * mm/page-discard.c
 *
 * (C) Copyright IBM Corp. 2005, 2006
 *
 * Guest page hinting functions.
 *
 * Authors: Martin Schwidefsky <schwidefsky@de.ibm.com>
 *          Hubertus Franke <frankeh@watson.ibm.com>
 *          Himanshu Raj <rhim@cc.gatech.edu>
 */

#include <linux/mm.h>
#include <linux/mm_inline.h>
#include <linux/pagemap.h>
#include <linux/rmap.h>
#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/buffer_head.h>
#include <linux/swap.h>

#include "internal.h"

/*
 * Check if there is something that prevents the page from changing
 * its state to volatile.
 */
static inline int __page_discardable(struct page *page,unsigned int offset)
{
	struct address_space *mapping;

	/*
	 * There are several conditions that prevent a page from becoming
	 * volatile. The first check is for the page bits.
	 */
	if (PageDirty(page) || PageReserved(page) || PageWriteback(page) ||
	    PageLocked(page) || PagePrivate(page) || PageDiscarded(page) ||
	    !PageUptodate(page) || !PageLRU(page) ||
	    (PageAnon(page) && !PageSwapCache(page)))
		return 0;

	/*
	 * If the page has been truncated there is no point in making
	 * it volatile. It will be freed soon. If the mapping ever had
	 * locked pages all pages of the mapping will stay stable.
	 */
	mapping = page_mapping(page);
	if (!mapping || mapping->mlocked)
		return 0;

	/*
	 * The last check is the critical one. We check the reference
	 * counter of the page against the number of mappings. The caller
	 * of make_volatile passes an offset, that is the number of extra
	 * references. For most calls that is 1 extra reference for the
	 * page-cache. In some cases the caller itself holds an additional
	 * reference, then the offset is 2. If the page map counter is equal
	 * to the page count minus the offset then there is no other
	 * (unknown) user of the page in the system and we can make the
	 * page volatile.
	 */
	if (page_mapcount(page) + offset != page_count(page))
		return 0;

	return 1;
}

/*
 * Attempts to change the state of a page to volatile. If there is
 * something preventing the state change the page stays in its current
 * state.
 */
void __page_make_volatile(struct page *page, unsigned int offset)
{
	/*
	 * If we can't get the PG_state_change bit just give up. The
	 * worst that can happen is that the page will stay in stable
	 * state although it might be volatile.
	 */
	preempt_disable();
	if (!TestSetPageStateChange(page)) {
		if (__page_discardable(page, offset))
			page_set_volatile(page, PageWritable(page));
		ClearPageStateChange(page);
	}
	preempt_enable();
}
EXPORT_SYMBOL(__page_make_volatile);

/*
 * Attempts to change the state of a page to stable. The host could
 * have removed a volatile page, the page_set_stable_if_present call
 * can fail.
 *
 * returns "0" on success and "1" on failure
 */
int __page_make_stable(struct page *page)
{
	/*
	 * Postpone state change to stable until PG_state_change bit is
	 * cleared. As long as PG_state_change is set another cpu is in
	 * page_make_volatile for this page. That makes sure
	 * that no caller of make_stable "overtakes" a make_volatile
	 * leaving the page in volatile where stable is required.
	 * The caller of make_stable need to make sure that no caller
	 * of make_volatile can make the page volatile right after
	 * make_stable has finished. That is done by requiring that
	 * page has been locked or that the page_count has been
	 * increased before make_stable is called. In both cases a
	 * subsequent call page_make_volatile will fail.
	 */
	while (PageStateChange(page))
		cpu_relax();
	return page_set_stable_if_present(page);
}
EXPORT_SYMBOL(__page_make_stable);

/**
 * __page_check_writable() - check page state for new writable pte
 *
 * @page: the page the new writable pte refers to
 * @pte: the new writable pte
 */
void __page_check_writable(struct page *page, pte_t pte, unsigned int offset)
{
	preempt_disable();
	while (TestSetPageStateChange(page))
		cpu_relax();

	if (!PageWritable(page)) {
		if (__page_discardable(page, offset))
			page_set_volatile(page, 1);
		else
			/*
			 * If two processes create a write mapping at the
			 * same time __page_discardable will return
			 * false but the page IS in volatile state.
			 * We have to take care about the dirty bit so the
			 * only option left is to make the page stable.
			 */
			page_set_stable_if_present(page);
		SetPageWritable(page);
	}
	ClearPageStateChange(page);
	preempt_enable();
}
EXPORT_SYMBOL(__page_check_writable);

/**
 * __page_reset_writable() - clear the PageWritable bit
 *
 * @page: the page
 */
void __page_reset_writable(struct page *page)
{
	preempt_disable();
	if (!TestSetPageStateChange(page)) {
		ClearPageWritable(page);
		ClearPageStateChange(page);
	}
	preempt_enable();
}
EXPORT_SYMBOL(__page_reset_writable);

/**
 * __page_discard() - remove a discarded page from the cache
 *
 * @page: the page
 *
 * The page passed to this function needs to be locked.
 */
static void __page_discard(struct page *page)
{
	struct address_space *mapping;
	struct zone *zone;

	/* Paranoia checks. */
	BUG_ON(PageWriteback(page));
	BUG_ON(PageDirty(page));
	BUG_ON(PagePrivate(page));

	/* Set the discarded bit early. */
	if (TestSetPageDiscarded(page))
		return;

	/* Unmap the page from all page tables. */
	page_unmap_all(page);

	/* Check if really all mappers of this page are gone. */
	BUG_ON(page_mapcount(page) != 0);

	/*
	 * Remove the page from LRU if it is currently added.
	 * The users of isolate_lru_pages need to check the
	 * discarded bit before readding the page to the LRU.
	 */
	zone = page_zone(page);
	spin_lock_irq(&zone->lru_lock);
	if (TestClearPageLRU(page)) {
		/* Unlink page from lru. */
		del_page_from_lru(zone, page);
	}
	spin_unlock_irq(&zone->lru_lock);

	/* Remove page from page cache/swap cache. */
 	mapping = page->mapping;
	if (PageSwapCache(page)) {
		swp_entry_t entry = { .val = page_private(page) };
		write_lock_irq(&swapper_space.tree_lock);
		__delete_from_swap_cache_nocheck(page);
		write_unlock_irq(&swapper_space.tree_lock);
		swap_free(entry);
		page_cache_release(page);
	} else {
		write_lock_irq(&mapping->tree_lock);
		__remove_from_page_cache_nocheck(page);
		write_unlock_irq(&mapping->tree_lock);
 		__put_page(page);
	}
}

/**
 * page_discard() - remove a discarded page from the cache
 *
 * @page: the page
 *
 * Before calling this function an additional page reference needs to
 * be acquired. This reference is released by the function.
 */
void page_discard(struct page *page)
{
	lock_page(page);
	__page_discard(page);
	unlock_page(page);
	page_cache_release(page);
}
EXPORT_SYMBOL(page_discard);
